/* Basic Access Control for ICAO-compliant MTRD's 
 *
 * As described in TR-PKI 1.1
 * 	Technical Report - PKI for Machine Readable Travel Documents offering
 * 	ICC Read-Only Access, Version 1.1, October 01, 2004.
 *
 * (C) 2005 by Harald Welte <laforge@hmw-consulting.de>
 *
 */ 

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <netinet/in.h>

#include <openssl/sha.h>
#include <openssl/rand.h>
#include <openssl/des.h>

#include <libmrtd/icao.h>
#include <libmrtd/libmrtd.h>

#include "iso7816_util.h"
#include "tlv.h"

/* If TEST is defined, we use test data from TR-PKI 1.1 */
/* #define TEST */

#ifdef TEST
static char test_bac_k_enc[] = { 
	0xab, 0x94, 0xfd, 0xec, 0xf2, 0x67, 0x4f, 0xdf,
	0xb9, 0xb3, 0x91, 0xf8, 0x5d, 0x7f, 0x76, 0xf2
};
static char test_bac_k_mac[] = { 
	0x79, 0x62, 0xD9, 0xEC, 0xE0, 0x3D, 0x1A, 0xCD, 
	0x4C, 0x76, 0x08, 0x9D, 0xCE, 0x13, 0x15, 0x43
};
static char test_rnd_icc[] = { 
	0x46, 0x08, 0xF9, 0x19, 0x88, 0x70, 0x22, 0x12
};
static char test_rnd_ifd[] = { 
	0x78, 0x17, 0x23, 0x86, 0x0C, 0x06, 0xC2, 0x26
};
static char test_k_ifd[] = {
	0x0B, 0x79, 0x52, 0x40, 0xCB, 0x70, 0x49, 0xB0,
	0x1C, 0x19, 0xB3, 0x3E, 0x32, 0x80, 0x4F, 0x0B
};
static char test_resp_apdu[] = {
	0x46, 0xB9, 0x34, 0x2A, 0x41, 0x39, 0x6C, 0xD7,
	0x38, 0x6B, 0xF5, 0x80, 0x31, 0x04, 0xD7, 0xCE,
	0xDC, 0x12, 0x2B, 0x91, 0x32, 0x13, 0x9B, 0xAF,
	0x2E, 0xED, 0xC9, 0x4E, 0xE1, 0x78, 0x53, 0x4F,
	0x2F, 0x2D, 0x23, 0x5D, 0x07, 0x4D, 0x74, 0x49,
	0x90, 0x00
};
static char test_sel_efcom_resp[] = {
	0x99, 0x02, 0x90, 0x00, 0x8E, 0x08, 0xFA, 0x85,
	0x5A, 0x5D, 0x4C, 0x50, 0xA8, 0xED, 0x90, 0x00
};
#endif


#if 0
#define DEBUGP(x, args ...)	fprintf(stderr, "%s(%d):%s: " x, __FILE__, __LINE__, __FUNCTION__, ## args)
#else
#define DEBUGP(x, args ...)
#endif

/* this is valid on little endian only. */
u_int64_t htonll(u_int64_t in)
{
	u_int64_t out;

	out = ((in & (u_int64_t)0x00000000000000ffULL) << 56) |
	      ((in & (u_int64_t)0x000000000000ff00ULL) << 40) |
	      ((in & (u_int64_t)0x0000000000ff0000ULL) << 24) |
	      ((in & (u_int64_t)0x00000000ff000000ULL) <<  8) |
	      ((in & (u_int64_t)0x000000ff00000000ULL) >>  8) |
	      ((in & (u_int64_t)0x0000ff0000000000ULL) >> 24) |
	      ((in & (u_int64_t)0x00ff000000000000ULL) >> 40) |
	      ((in & (u_int64_t)0xff00000000000000ULL) >> 56);

	return out;
}
#define ntohll htonll


DES_cblock zero_ivec;			/* zero */

static void xor(unsigned char *dst, const unsigned char *a, 
		const unsigned char *b, unsigned int len)
{
	int i;

	/* this could be unrolled */
	for (i = 0; i < len; i++)
		dst[i] = a[i] ^ b[i];
}

/* length without checkdigit */
#define ICAO_MRZ_DOCNR_LEN	9
#define ICAO_MRZ_DOB_LEN	6
#define ICAO_MRZ_DOE_LEN	6

/* generates "MRZ_information" (Chapter 3.2.2, TR-PKI 1.0) */
static int 
mrz_to_mrzinfo(unsigned char *mrzinfo, const char *mrz2)
{
	char doc_nr[ICAO_MRZ_DOCNR_LEN+2];
	char date_of_birth[ICAO_MRZ_DOB_LEN+2];
	char date_of_expiry[ICAO_MRZ_DOE_LEN+2];

	if (strlen(mrz2) < 28)
		return -1;

	strncpy(doc_nr, mrz2, 10);
	doc_nr[10] = '\0';

	strncpy(date_of_birth, mrz2+13, 7);
	date_of_birth[7] = '\0';

	strncpy(date_of_expiry, mrz2+21, 7);
	date_of_expiry[7] = '\0';

	*mrzinfo = '\0';
	strncat((char *)mrzinfo, doc_nr, ICAO_MRZ_DOCNR_LEN+1);
	strncat((char *)mrzinfo, date_of_birth, ICAO_MRZ_DOB_LEN+1);
	strncat((char *)mrzinfo, date_of_expiry, ICAO_MRZ_DOE_LEN+1);
	DEBUGP("mrzinfo ='%s'\n", mrzinfo);

	return 0;
}

/* Pad a data to a block length of 8, filling with 0x80 0x00 .... */
static int
sm_pad(unsigned char *dst, const unsigned char *src, unsigned int orig_len)
{
	unsigned int blocks = (orig_len / 8) + 1;

	memcpy(dst, src, orig_len);
	dst[orig_len] = 0x80;
	memset(dst+orig_len+1, 0, blocks*8 - (orig_len+1));

	return blocks*8;
}

/* retail MAC as described in CWA 14890-1, Sec. 9.7 (SSC in net byte order) */
static int
des_retail_mac(DES_cblock *cc, const_DES_cblock *ssc, 
	       DES_key_schedule *ka, DES_key_schedule *kb,
	       const unsigned char *data, unsigned int len)
{
	DES_cblock yN;
	unsigned int rounds;
	int i;
	unsigned char *padbuf = malloc(len + 8);
	unsigned int padlen;

	if (!padbuf)
		return -ENOMEM;

	DEBUGP("SSC=0x%p=[%s]\n", ssc, !ssc ? NULL : mrtd_hexdump(ssc, 8));
	DEBUGP("Data=%s\n", mrtd_hexdump(data, len));

	/* pad */
	padlen = sm_pad(padbuf, data, len);

	if (!ssc) {
		/* special case: ICB is zero during key generation */
		memset(&yN, 0, sizeof(yN));
	} else {
		/* first step */
		DES_ecb_encrypt(ssc, &yN, ka, DES_ENCRYPT);
	}

	/* second step(s) */
	rounds = padlen / 8;
	for (i = 0; i < rounds; i++) {
		u_int64_t *cur_msb = (u_int64_t *) (padbuf + i*8);
		*(u_int64_t *) yN ^= *cur_msb;
		DES_ecb_encrypt(&yN, &yN, ka, DES_ENCRYPT);
	}
	
	/* final 3des phase */
	DES_ecb_encrypt(&yN, &yN, kb, DES_DECRYPT);
	DES_ecb_encrypt(&yN, cc, ka, DES_ENCRYPT);
	DEBUGP("CC=%s\n", mrtd_hexdump(cc, 8));

	free(padbuf);
	return 0;
}

static int
esign_compute_session_keys(unsigned char *k_enc,
			   unsigned char *k_mac,
			   unsigned const char *seed,
			   unsigned int seed_len)
{
#define MAX_KLEN	(2048/8)	/* according Sec. 8.8 CWA 14890-1 */
#define CTR_LEN		4		/* 32bit counter */
	unsigned char hash1_src[MAX_KLEN+CTR_LEN];
	unsigned char hash2_src[MAX_KLEN+CTR_LEN];
	unsigned char hash1[SHA_DIGEST_LENGTH];
	unsigned char hash2[SHA_DIGEST_LENGTH];

	if (seed_len > MAX_KLEN)
		seed_len = MAX_KLEN;

	memcpy(hash1_src, seed, seed_len);
	*((u_int32_t *) (hash1_src+seed_len)) = htonl(0x00000001);/* c = 1 */

	memcpy(hash2_src, seed, seed_len);
	*((u_int32_t *) (hash2_src+seed_len)) = htonl(0x00000002);/* c = 2 */

	SHA1((unsigned char *)hash1_src, seed_len+CTR_LEN,
	     (unsigned char *)hash1);

	SHA1((unsigned char *)hash2_src, seed_len+CTR_LEN,
	     (unsigned char *)hash2);

	memcpy(k_enc, hash1, 16);
	memcpy(k_mac, hash2, 16);

	/* Set the parity to create valid 3des key pair */
	DES_fixup_key_parity((DES_cblock *)k_enc);
	DES_fixup_key_parity((DES_cblock *)(k_enc+DES_KEY_SZ));

	/* Set the parity to create valid 3des key pair */
	DES_fixup_key_parity((DES_cblock *)k_mac);
	DES_fixup_key_parity((DES_cblock *)(k_mac+DES_KEY_SZ));

	return 0;
}

#define KEY_SEED_LEN	(2*DES_KEY_SZ)

static int
calc_bac_keys(unsigned char *bac_k_enc,
	      unsigned char *bac_k_mac, const char *mrz2)
{
	unsigned char mrzinfo[ICAO_MRZ_DOCNR_LEN+ICAO_MRZ_DOB_LEN+ICAO_MRZ_DOE_LEN+3];
	unsigned char sha1_mrzinfo[SHA_DIGEST_LENGTH];
	unsigned char key_seed[KEY_SEED_LEN];
	int rv;

	rv = mrz_to_mrzinfo(mrzinfo, mrz2);
	if (rv < 0)
		return rv;

	/* calculate sha-1 hash */
	SHA1((unsigned char *)mrzinfo, sizeof(mrzinfo),
	     (unsigned char *)sha1_mrzinfo);
	DEBUGP("SHA1mrzinfo =%s\n", mrtd_hexdump(sha1_mrzinfo,
						 sizeof(sha1_mrzinfo)));

	memcpy(key_seed, sha1_mrzinfo, KEY_SEED_LEN);
	DEBUGP("Kseed =%s\n", mrtd_hexdump(key_seed, sizeof(key_seed)));

	rv = esign_compute_session_keys(bac_k_enc, 
					bac_k_mac, key_seed, KEY_SEED_LEN);
	if (rv < 0)
		return rv;
	DEBUGP("Kenc =%s\n", mrtd_hexdump(bac_k_enc, 16));
	DEBUGP("Kmac =%s\n", mrtd_hexdump(bac_k_mac, 16));

	return 0;
}

/* receive/verify/decrypt an RAPDU from the ICC */
static int
bac_transceive_response(struct mrtd_handle *h,
			const unsigned char *resp, unsigned int resp_len,
			unsigned char *rx, unsigned int *rx_len)
{
	struct tlv t, t_do87, t_do99, t_do8e;
	unsigned int do87_present = 0;
	unsigned int do99_present = 0;
	unsigned int do8e_present = 0;
	const unsigned char *cur = resp;
	unsigned char *cur2;
	unsigned char sw[2];
	int ret = 0;
	DES_cblock cc;
	u_int64_t ssc_net;

	unsigned int k_len;
	unsigned char *k = NULL;

	memset(rx, 0, *rx_len);

	DEBUGP("RAPDU =%s\n", mrtd_hexdump(resp, resp_len));

	do {
		if (cur + 2 == resp + resp_len) {
			sw[0] = *cur;
			sw[1] = *cur+1;
			break;
		} 
		if (tlv_parse(cur, resp_len - (cur - resp), &t) < 0) {
			fprintf(stderr, "Error parsing TLV\n");
			ret = -EINVAL;
			goto out;
		}

		switch (tlv_raw_tag(&t)) {
		case 0x87:
			/* encrypted data */
			memcpy(&t_do87, &t, sizeof(t));
			do87_present = 1;
			DEBUGP("DO87 [%u] = %s\n", t_do87.len,
				mrtd_hexdump(t_do87.val, t_do87.len));
			break;
		case 0x99:
			/* status words */
			memcpy(&t_do99, &t, sizeof(t));
			do99_present = 1;
			DEBUGP("DO99 [%u] = %s\n", t_do99.len,
				mrtd_hexdump(t_do99.val, t_do99.len));
			break;
		case 0x8e:
			/* cryptographic checksum */
			memcpy(&t_do8e, &t, sizeof(t));
			do8e_present = 1;
			DEBUGP("DO8E [%u] = %s\n", t_do8e.len,
				mrtd_hexdump(t_do8e.val, t_do8e.len));
			break;
		}

		cur += tlv_gross_len(&t);
	} while (cur < resp + resp_len);

	if (!do99_present || !do8e_present) {
		fprintf(stderr, "DO99 and/or DO8E missing\n");
		if (resp_len == 2 && *rx_len >= 2) {
			memcpy(rx, resp, resp_len);
			*rx_len = resp_len;
		} else
			ret = -EINVAL;
		goto out;
	}

	/* Increment SSC */
	h->bac.sstate.ssc++;
	ssc_net = htonll(h->bac.sstate.ssc);

	/* concatenate and pad SSC, DO87 and DO99 */
	k_len = tlv_gross_len(&t_do99) + sizeof(DES_cblock);
	if (do87_present)
		k_len += tlv_gross_len(&t_do87);
	k = malloc(k_len);
	if (!k) {
		fprintf(stderr, "ENOMEM\n");
		ret = -ENOMEM;
		goto out_free;
	}

	k_len = 0;
	if (do87_present) {
		memcpy(k, t_do87._tagptr, tlv_gross_len(&t_do87));
		k_len += tlv_gross_len(&t_do87);
	}

	memcpy(k+k_len, t_do99._tagptr, tlv_gross_len(&t_do99));
	k_len += tlv_gross_len(&t_do99);

	ret = des_retail_mac(&cc, (DES_cblock *)&ssc_net,
			     &h->bac.sstate.ks_mac_ks1,
			     &h->bac.sstate.ks_mac_ks2, k, k_len);
	if (ret < 0) {
		fprintf(stderr, "Error computing MAC\n");
		goto out_free;
	}
	DEBUGP("CC =%s\n", mrtd_hexdump(&cc, sizeof(cc)));

	if (memcmp(cc, t_do8e.val, sizeof(cc))) {
		fprintf(stderr, "MAC of received response wrong\n");
		ret = -EINVAL;
		goto out_free;
	}

	cur2 = rx;
	if (do87_present) {
		unsigned int rx_len_unpadded = t_do87.len-1;

		if (*t_do87.val != 0x01) {
			fprintf(stderr, "unknown padding scheme 0x%x\n",
				*t_do87.val);
			ret = -EIO;
			goto out_free;
		}

		/* Decrypt data of DO87 */
		memset(&zero_ivec, 0, sizeof(zero_ivec));
		DES_ede2_cbc_encrypt((unsigned char *)t_do87.val+1,
				     (unsigned char *)rx, 
				     t_do87.len-1, &h->bac.sstate.ks_enc_ks1, 
				     &h->bac.sstate.ks_enc_ks2,
				     &zero_ivec, DES_DECRYPT);

		/* Determine unpadded length */
		while (rx_len_unpadded >= 1 && rx[rx_len_unpadded-1] == 0x00)
			rx_len_unpadded--;

		if (rx_len_unpadded >= 1 && rx[rx_len_unpadded-1] == 0x80)
			rx_len_unpadded--;
		else
			rx_len_unpadded = t_do87.len-1;

		if (rx_len_unpadded + 2 > *rx_len) {
			fprintf(stderr, "rx_len(%u) short DO87(%u-%u)!\n",
				*rx_len, t_do87.len, rx_len_unpadded);
			ret = -EINVAL;
			goto out_free;
		}

		cur2 += rx_len_unpadded;

	}

	/* copy status word at end of response */
	*(cur2++) = t_do99.val[0];
	*(cur2++) = t_do99.val[1];

	*rx_len = (cur2 - rx);

	DEBUGP("rx [%u] =%s\n", *rx_len, mrtd_hexdump(rx, *rx_len));

out_free:
	free(k);
out:
	return ret;
}



/* transceive function for SM en-/decapsulation */
static int
bac_transceive(struct mrtd_handle *h, 
	       const unsigned char *tx, unsigned int tx_len,
	       unsigned char *rx, unsigned int *rx_len)
{
	unsigned char cmd_hdr[4 + 4]; /* CLA INS P1 P2  +  padding */
	unsigned int lc = 0, le = 0;
	const unsigned char *cmd_data;
	unsigned char *cur;

	unsigned int do87_len = 0;
	unsigned char *do87 = NULL;
	unsigned int do97_len = 0;
	unsigned char *do97 = NULL;
	DES_cblock cc;
	unsigned char do8e[2 + sizeof(DES_cblock)];

	unsigned char *apdu = NULL;
	unsigned int apdu_lc, apdu_len;

	unsigned char *n = NULL;

	unsigned int m_len;
	unsigned char *m = NULL;

	unsigned char resp[32768];
	unsigned int resp_len = sizeof(resp);

	u_int64_t ssc_net;
	int rv, ret;

	if (tx_len < 4)
		return -EINVAL;

	/* Copy + Pad command header */
	sm_pad(cmd_hdr, tx, 4);

	/* Mask Class byte according to 7816-4 5.4.1 */
	cmd_hdr[0] = (cmd_hdr[0] & 0xf0) | 0x0c;
	DEBUGP("cmd_hdr =%s\n", mrtd_hexdump(cmd_hdr, sizeof(cmd_hdr)));

	/* Parse APDU */
	if (tx_len == 4) {
		/* Case 1 */
		lc = 0;
		le = -1;
		cmd_data = NULL;
	} else if (tx_len == 5) {
		/* Case 2: Short APDU */
		lc = 0;
		le = tx[4];
		cmd_data = NULL;
	} else if (tx_len == 5 + tx[4] && tx[4] != 0) {
		/* Case 3S */
		lc = tx[4];
		le = -1;
		cmd_data = &tx[5];
	} else if (tx_len == 6 + tx[4] && tx[4] != 0) {
		/* Case 4S */
		lc = tx[4];
		le = tx[tx_len-1];
		cmd_data = &tx[5];
	} else if (tx_len == 7 && tx[4] == 0) {
		/* Case 2E */
		lc = 0;
		le = (tx[5] << 8) | tx[6];
		cmd_data = NULL;
	} else {
		lc = (tx[5] << 8) | tx[6];
		cmd_data = &tx[7];
		if (tx_len == 7 + lc && tx[4] == 0 && lc != 0) {
			/* Case 3E */
			le = -1;
		} else if (tx_len == 9 + lc && tx[4] == 0 && lc != 0) {
			/* Case 4E */
			le = (tx[tx_len-2] << 8 | tx[tx_len-1]);
		} else
			return -EINVAL;
	}

	//if (lc) {
	if (1) {
		/* Build DO87 */
		unsigned char *do87_src = malloc(lc + 1);
		unsigned int do87_payload_ofs;
		unsigned int do87_src_len;

		if (!do87_src)
			return -ENOMEM;

 		do87 = malloc(lc + sizeof(DES_cblock));
		if (!do87) {
			ret = -ENOMEM;
			goto out_free;
		}

		/* copy and pad command data */
		do87_src_len = sm_pad(do87_src, cmd_data, lc);
		DEBUGP("Data =%s\n", mrtd_hexdump(do87_src, do87_src_len));

		/* prefix with 87, length, 01 */
		do87_payload_ofs = tlv_build_tl(do87, 0x87, do87_src_len+1);
		*(do87 + do87_payload_ofs) = 0x01;
		DEBUGP("TLV87 =%s\n", mrtd_hexdump(do87, do87_src_len+1));

		/* encrypt with 3des */
		memset(&zero_ivec, 0, sizeof(zero_ivec));
		DES_ede2_cbc_encrypt((unsigned char *)do87_src, 
				     (unsigned char *)do87+do87_payload_ofs+1, 
				     do87_src_len, &h->bac.sstate.ks_enc_ks1, 
				     &h->bac.sstate.ks_enc_ks2,
				     &zero_ivec, DES_ENCRYPT);

		do87_len = do87_src_len + do87_payload_ofs + 1;
		DEBUGP("DO87 =%s\n", mrtd_hexdump(do87, do87_len));

		free(do87_src);
	}

	if (le != -1) {
		/* Build DO97 */
		unsigned int le_len, ofs;
		do97 = malloc(le_len+8);
		if (!do97) {
			ret = -ENOMEM;
			goto out_free;
		}

		if (le <= 0xff)
			le_len = 1;
		else
			le_len = 2;

		ofs = tlv_build_tl(do97, 0x97, le_len);

		if (le_len == 2) {
			*(do97+ofs) = le >> 8;
			*(do97+ofs+1) = le & 0xff;
			do97_len = ofs + 2;
		} else {
			*(do97+ofs) = le & 0xff;
			do97_len = ofs + 1;
		}
		DEBUGP("DO97 =%s\n", mrtd_hexdump(do97, do97_len));
	}

	/* concatenate cmd_hdr and DO87 (if any) and DO 97 (if any)*/
	m_len = sizeof(cmd_hdr) + do87_len + do97_len;
	m = malloc(m_len);
	if (!m)
		goto out_free;
	memcpy(m, cmd_hdr, sizeof(cmd_hdr));
	if (do87)
		memcpy(m+sizeof(cmd_hdr), do87, do87_len);
	if (do97)
		memcpy(m+sizeof(cmd_hdr)+do87_len, do97, do97_len);

	DEBUGP("M =%s\n", mrtd_hexdump(m, m_len));

	/* increment SSC */
	h->bac.sstate.ssc++;
	ssc_net = htonll(h->bac.sstate.ssc);

	/* compute MAC over N with KSmac */
	des_retail_mac(&cc, (const_DES_cblock *)&ssc_net,
			&h->bac.sstate.ks_mac_ks1,
			&h->bac.sstate.ks_mac_ks2, m, m_len);

	/* build DO8E */
	do8e[0] = 0x8e;
	do8e[1] = sizeof(DES_cblock);
	memcpy(do8e+2, cc, sizeof(DES_cblock));

	/* construct protected APDU */
	apdu_lc = do87_len + do97_len + sizeof(do8e);
	apdu_len = apdu_lc + 4 + 1 + 1;	/* lc + hdr_len + len + 1 byte at end */
	apdu = malloc(apdu_len);
	if (!apdu) {
		ret = -ENOMEM;
		goto out_free;
	}

	/* command header */
	memcpy(apdu, cmd_hdr, 4);

	/* length of command */
	apdu[4] = apdu_lc & 0xff;
	cur = apdu+5;
	/* add do87 */
	memcpy(cur, do87, do87_len);
	cur += do87_len;
	if (do97) {
		/* add do97 */
		memcpy(cur, do97, do97_len);
		cur += do97_len;
	}
	/* add do8e */
	memcpy(cur, do8e, sizeof(do8e));
	cur += sizeof(do8e);
	/* Fake Le */
	*cur = 0x00;
	DEBUGP("ProtectedAPDU =%s\n", mrtd_hexdump(apdu, apdu_len));

#ifdef TEST
	resp_len = sizeof(test_sel_efcom_resp);
	memcpy(resp, test_sel_efcom_resp, resp_len);
#else
	rv = h->bac.transceive(h, apdu, apdu_len, resp, &resp_len);
	if (rv < 0) {
		fprintf(stderr, "low-layer transceive failed\n");
		goto out_free;
	}
#endif
	ret = bac_transceive_response(h, resp, resp_len, rx, rx_len);

out_free:
	if (apdu)
		free(apdu);
	if (m)
		free(m);
	if (n)
		free(n);
	if (do87)
		free(do87);
	if (do97)
		free(do97);
	return ret;
}
			
int 
basic_authenticate(struct mrtd_handle *h, char *mrz2)
{
	/* what a stack... */
	unsigned char bac_k_enc[KEY_LEN_TDES];
	unsigned char bac_k_mac[KEY_LEN_TDES];	/* basic access control */

	u_int64_t rnd_icc;	/* RND.ICC */
	u_int64_t rnd_ifd;	/* RND.IFD */

	unsigned char k_icc[DES_KEY_SZ*2];	/* Kicc */
	unsigned char k_ifd[DES_KEY_SZ*2];	/* Kifd */
	unsigned char k_ifd_icc[DES_KEY_SZ*2];	/* Kifd/icc */

	unsigned char s[sizeof(rnd_icc)+sizeof(rnd_ifd)+sizeof(k_ifd)];
	unsigned char enc_s[sizeof(s)];	/* encrypted seed */
	unsigned char mac_s[8];		/* mac of encrypted seed */

	unsigned char cmd[5+sizeof(enc_s)+sizeof(mac_s)+1];

	unsigned char r[sizeof(rnd_icc)+sizeof(rnd_ifd)+sizeof(k_icc)];
	unsigned char enc_r[sizeof(r)];
	unsigned char mac_r[8];
	unsigned int enc_r_len = sizeof(enc_r);

	DES_key_schedule k_enc_ks1, k_enc_ks2;	/* for bac_k_enc */
	DES_key_schedule k_mac_ks1, k_mac_ks2;	/* for bac_k_mac */
	DES_cblock zero_ivec;			/* zero */

	int rv;

	memset(zero_ivec, 0, sizeof(zero_ivec));

	/* initialize to zero for padding */
	memset(s, 0, sizeof(s));

#ifdef TEST
	memcpy(bac_k_enc, test_bac_k_enc, sizeof(bac_k_enc));
	memcpy(bac_k_mac, test_bac_k_mac, sizeof(bac_k_mac));
#else
	rv = calc_bac_keys(bac_k_enc, bac_k_mac, mrz2);
	if (rv < 0) {
		fprintf(stderr, "error calculating BAC keys\n");
		return rv;
	}
#endif

#ifdef TEST
	memcpy(&rnd_icc, test_rnd_icc, sizeof(rnd_icc));
#else
	rv = iso7816_get_challenge(h, (char *)&rnd_icc, sizeof(rnd_icc));
	if (rv < 0) {
		fprintf(stderr, "error getting challenge from card\n");
		return rv;
	}
#endif

	/* generate random rnd_ifd, k_ifd */
#ifdef TEST
	memcpy(&rnd_ifd, test_rnd_ifd, sizeof(rnd_ifd));
#else
	rv = RAND_bytes((unsigned char *)&rnd_ifd, sizeof(rnd_ifd));
	if (rv != 1)
		return -1;
#endif
		
#ifdef TEST
	memcpy(k_ifd, test_k_ifd, sizeof(k_ifd));
#else
	rv = RAND_bytes((unsigned char *)k_ifd, sizeof(k_ifd));
	if (rv != 1)
		return -1;
#endif

	/* build the seed (S = rnd_ifd || rnd_icc || k_ifd */
	memcpy(s, &rnd_ifd, sizeof(rnd_ifd));
	memcpy(s+sizeof(rnd_ifd), &rnd_icc, sizeof(rnd_icc));
	memcpy(s+sizeof(rnd_ifd)+sizeof(rnd_icc), k_ifd, sizeof(k_ifd));
	DEBUGP("S = %s\n", mrtd_hexdump(s, sizeof(s)));

	/* encrypt seed */
	DES_set_key_unchecked((const_DES_cblock *) bac_k_enc, &k_enc_ks1);
	DES_set_key_unchecked((const_DES_cblock *) (bac_k_enc+8), &k_enc_ks2);
	memset(&zero_ivec, 0, sizeof(zero_ivec));
	DES_ede2_cbc_encrypt((unsigned char *)s, (unsigned char *)enc_s, 
			     sizeof(s), &k_enc_ks1, &k_enc_ks2,
			     &zero_ivec, DES_ENCRYPT);
	DEBUGP("Eifd = %s\n", mrtd_hexdump(enc_s, sizeof(s)));

	/* mac of encrypted seed */
	DEBUGP("Kmac = %s\n", mrtd_hexdump(bac_k_mac, 16));
	DES_set_key_unchecked((const_DES_cblock *) bac_k_mac, &k_mac_ks1);
	DES_set_key_unchecked((const_DES_cblock *) (bac_k_mac+8), &k_mac_ks2);
	des_retail_mac((DES_cblock *)mac_s, NULL, 
		       &k_mac_ks1, &k_mac_ks2, enc_s, sizeof(enc_s));
	DEBUGP("Mifd = %s\n", mrtd_hexdump(mac_s, sizeof(mac_s)));

	cmd[0] = 0x00;
	cmd[1] = 0x82;	/* MUTUAL AUTHENTICATE */
	cmd[2] = 0x00;
	cmd[3] = 0x00;	/* key index 0 ? */
	cmd[4] = 0x28; /* lc */
	memcpy(cmd+5, enc_s, sizeof(enc_s));
	memcpy(cmd+5+sizeof(enc_s), mac_s, sizeof(mac_s));
	cmd[5+sizeof(enc_s)+sizeof(mac_s)] = 0x28;

	DEBUGP("cmd_data = %s\n", 
	       mrtd_hexdump(cmd, 6+sizeof(enc_s)+sizeof(mac_s)));

#ifdef TEST
	memcpy(enc_r, test_resp_apdu, sizeof(test_resp_apdu));
	enc_r_len = sizeof(test_resp_apdu);
#else
	/* R = rnd_icc || rnd_ifd || k_icc  ( 8 + 8 + 32) */
	/* we expeect 56bytes return (48byte encrypted R; 8byte MAC) */
	rv = h->fn.transceive(h, cmd, 6+sizeof(enc_s)+sizeof(mac_s),
			      enc_r, &enc_r_len);
	if (rv < 0) {
		fprintf(stderr, "error transcieving MUTUAL AUTH\n");
		return rv;
	}
#endif
	DEBUGP("resp_apdu = %s\n", mrtd_hexdump(enc_r, enc_r_len));

	/* check MAC */
	des_retail_mac((DES_cblock *)mac_r, NULL, 
			&k_mac_ks1, &k_mac_ks2, enc_r, 32);
	if (memcmp(mac_r, enc_r+32, 8)) {
		fprintf(stderr, "MAC from MRTD wrong\n");
		return -1;
	}
	DEBUGP("MAC correct\n");

	/* decrypt data */
	memset(&zero_ivec, 0, sizeof(zero_ivec));
	DES_ede2_cbc_encrypt((unsigned char *)enc_r, (unsigned char *)r, 
			     32, &k_enc_ks1, &k_enc_ks2, &zero_ivec,
			     DES_DECRYPT);
	DEBUGP("r = %s\n", mrtd_hexdump(r, 32));

	/* check whether rnd_ifd/icc is the one we sent earlier */
	if (memcmp(&rnd_ifd, r+sizeof(rnd_icc), sizeof(rnd_ifd))) {
		fprintf(stderr, "rnd_ifd (0x%llx != 0x%llx)\n",
			rnd_ifd, *(u_int64_t *)(r+sizeof(rnd_icc)));
		return -1;
	}
	DEBUGP("RND.IFD correct\n");
			
	if (memcmp(&rnd_icc, r, sizeof(rnd_icc))) {
		fprintf(stderr, "rnd_icc(0x%llx != 0x%llx)\n",
			rnd_icc, *(u_int64_t *)r);
		return -1;
	}
	DEBUGP("RND.ICC correct\n");

	/* extract k_icc */
	memcpy(k_icc, r+sizeof(rnd_icc)+sizeof(rnd_ifd), sizeof(k_icc));

	/* calculate session keys */
	xor(k_ifd_icc, k_ifd, k_icc, sizeof(k_ifd_icc));
	DEBUGP("Kseed = %s\n", mrtd_hexdump(k_ifd_icc, sizeof(k_ifd_icc)));
	esign_compute_session_keys((unsigned char *)h->bac.sstate.ks_enc, 
				   (unsigned char *)h->bac.sstate.ks_mac, 
				   k_ifd_icc, sizeof(k_ifd_icc));
	DEBUGP("KSenc = %s\n", mrtd_hexdump(h->bac.sstate.ks_enc,
		sizeof(h->bac.sstate.ks_enc)));
	DEBUGP("KSmac = %s\n", mrtd_hexdump(h->bac.sstate.ks_mac,
		sizeof(h->bac.sstate.ks_mac)));

	/* build keyschedule from session keys */
	DES_set_key_unchecked((const_DES_cblock *) h->bac.sstate.ks_enc, 
			      &h->bac.sstate.ks_enc_ks1);
	DES_set_key_unchecked((const_DES_cblock *) (h->bac.sstate.ks_enc+8), 
			      &h->bac.sstate.ks_enc_ks2);
	DES_set_key_unchecked((const_DES_cblock *) h->bac.sstate.ks_mac, 
			      &h->bac.sstate.ks_mac_ks1);
	DES_set_key_unchecked((const_DES_cblock *) (h->bac.sstate.ks_mac+8), 
			      &h->bac.sstate.ks_mac_ks2);
	
	/* calculate initial SSC */
	h->bac.sstate.ssc = 
			(ntohll(rnd_icc) & 0xffffffff) << 32 | 
			(ntohll(rnd_ifd) & 0xffffffff);
	DEBUGP("SSC = (%s)\n", mrtd_hexdump(&h->bac.sstate.ssc, 8));

	/* overload transceive function */
	h->bac.transceive = h->fn.transceive;
	h->fn.transceive = &bac_transceive;

	return 0;
}
