/* libicao - communication with ICAO compliant MRTD's 
 *
 * Implementation of an ISO/IEC 19794-5 parser
 *
 * (C) 2005 by Harald Welte <hwelte@hmw-consulting.de>
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>

#include <libmrtd/iso19794_5.h>

struct iso19794_5_info *
iso19794_5_parse(const char *data, unsigned int len)
{
	struct iso19794_5_facial_record_hdr *frh = 
			(struct iso19794_5_facial_record_hdr *) data;
	struct iso19794_5_info *info;
	unsigned int info_size;
	unsigned int num_images = ntohs(frh->num_images);
	const unsigned char *cur;
	int i;

	if (!frh->format_id[3] == 0 || !frh->version[3] == 0) {
		errno = -EINVAL;
		return NULL;
	}

	if (strcmp(frh->format_id, ISO19794_5_FORMAT_IDENTIFIER)) {
		errno = -EINVAL;
		return NULL;
	}

	if (strcmp(frh->version, ISO19794_5_VERSION)) {
		errno = -EINVAL;
		return NULL;
	}

	if (ntohl(frh->length) > len) {
		errno = -EINVAL;
		return NULL;
	}

	info_size = sizeof(*info) +
		      (num_images * sizeof(struct mrtd_facial_record));
	info = malloc(info_size);
	if (!info) {
		errno = -ENOMEM;
		return NULL;
	}
	memset(info, 0, info_size);

	memcpy(info->frh.format_id, frh->format_id, 
		sizeof(info->frh.format_id));
	memcpy(info->frh.version, frh->version, sizeof(info->frh.version));
	info->frh.length = ntohl(frh->length);
	info->frh.num_images = num_images;

	cur = data + sizeof(*frh);

	for (i = 0; i < num_images; i ++) {
		const struct iso19794_5_facial_info_block *fib = 
			(struct iso19794_5_facial_info_block *) cur;
		struct mrtd_facial_record *fr = &info->fr[i];
		struct iso19794_5_image_info *iib;
		const unsigned char *orig_cur = cur;
		int j;

		memcpy(&fr->facial_info, fib, sizeof(fr->facial_info));
		/* convert multibyte value endianness */
		fr->facial_info.fi_block_length = ntohl(fib->fi_block_length);
		fr->facial_info.num_feature_points = 
						ntohl(fib->num_feature_points);
		fr->facial_info.expression = ntohs(fib->expression);

		cur += sizeof(*fib);
		for (j = 0; j < fr->facial_info.num_feature_points; j++) {
#if 0
			const struct iso19794_5_feature_point *fp = 
				(struct iso19794_5_feature_point *)cur;

			/* FIXME: parse feature point(s) */
#endif

			cur += sizeof(struct iso19794_5_feature_point);
		}

		iib = (struct iso19794_5_image_info *)cur;
		memcpy(&fr->image_info, iib, sizeof(&fr->image_info));
		/* convert multibyte value endianness */
		fr->image_info.width = ntohs(iib->width);
		fr->image_info.height = ntohs(iib->height);
		fr->image_info.device_type = ntohs(iib->device_type);
		fr->image_info.quality = ntohs(iib->quality);

		cur += sizeof(struct iso19794_5_image_info);
		fr->image_data = (char *)cur;
		fr->image_data_len = 
			(orig_cur + fr->facial_info.fi_block_length) - cur;

		cur = orig_cur + fr->facial_info.fi_block_length;
	}
		
	return info;
}

static char *gender_strings[] = {
	"unspecified",
	"male",
	"female",
	"unknown",
};

char *
iso19794_5_gender(u_int8_t gender)
{
	//if (gender > ARRAY_SIZE(gender_strings));
	return gender_strings[gender];
}

static char *eye_strings[] = {
	"unspecified",
	"black",
	"blue",
	"brown",
	"gray",
	"green",
	"multicolour",
	"pink",
	"other",
};

char *iso19794_5_eye_colour(u_int8_t colour)
{
	return eye_strings[colour];
}

static char *hair_strings[] = {
	"unspecified",
	"bald",
	"black",
	"blonde",
	"brown",
	"gray",
	"white",
	"red",
	"green",
	"blue",
	"other",
};

char *iso19794_5_hair_colour(u_int8_t colour)
{
	return hair_strings[colour];
}

static char *feature_strings[] = {
	"",
	"glasses",
	"moustache",
	"beard",
	"teeth visible",
	"blink",
	"mouth open",
	"left eye patch",
	"right eye patch",
	"dark (medical) glasses",
	"feature distorting medical condition",
};

char *iso19794_5_features(u_int32_t features)
{
	static char feature_buf[512];
	int i;

	features &= 0x7f;

	if (!(features & ISO19794_5_FEATURE_SPECIFIED))
		return "Unspecified";

	feature_buf[0] = '\0';

	for (i = 1; i < 24; i ++) {
		if (features & (1 << i)) {
			strcat(feature_buf, feature_strings[i]);
			strcat(feature_buf, ", ");
		}
	}
		
	return feature_buf;
}

static char *expression_strings[] = {
	"unspecified",
	"neutral",
	"closed smile",
	"open smile",
	"raised eyebrow",
	"looking away",
	"squinting",
	"frowning",
};

char *iso19794_5_expression(u_int16_t expression)
{
	if (expression & ISO19794_5_EXPR_VENDOR_MASK)
		return "vendor specified";
	return expression_strings[expression];
}

static char *imgtype_strings[] = {
	"unspecified",
	"basic",
	"full front",
	"token front",
	"other",
};

char *iso19794_5_img_type(u_int8_t type)
{
	return imgtype_strings[type];
}

static char *dtype_strings[] = {
	"JPEG",
	"JPEG2000",
};

char *iso19794_5_img_data_type(u_int8_t dtype)
{
	return dtype_strings[dtype];
}

static char *cspace_strings[] = {
	"unspecified",
	"RGB24",
	"YUV422",
	"8bit Greyscale",
	"other",
};

char *iso19794_5_img_cspace(u_int8_t cspace)
{
	return cspace_strings[cspace];
}

static char *source_strings[] = {
	"unspecified",
	"static image from unspecified source",
	"static image from digital camera",
	"static image from scanner",
	"video still frame from unknown source",
	"video still frame from analog source",
	"video still frame from digital source",
	"unknown",
};

char *iso19794_5_img_source(u_int8_t source)
{
	return source_strings[source];
}
