#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>

#include "tlv.h"

#define TLV_CLASS_MASK		0xc0
#define	TLV_CLASS_UNIVERSAL	0x00
#define TLV_CLASS_APPLICATION	0x40
#define TLV_CLASS_CONTEXT	0x80
#define TLV_CLASS_PRIVATE	0xc0

#define TLV_TYPE_COMBINED	0x20
#define TLV_TYPE_TAG_MASK	0x1f 

static int tlv_determine_tag(struct tlv *t)
{
	unsigned char *cur = t->_tagptr;

	if ((*cur & TLV_TYPE_TAG_MASK) == TLV_TYPE_TAG_MASK) {
		/* we only support tag lengths of up to two octets */
		if (t->_len < 2 || (*(cur+1) & 0x80))
			return -EINVAL;
		t->tag = (*(cur+1) & 0x7f);
		t->tag |= (*cur & TLV_TYPE_TAG_MASK) << 7;
		//t->tag = ntohs(t->tag);
		cur += 2;
	} else {
		t->tag = *cur & TLV_TYPE_TAG_MASK;
		cur += 1;
	}

	t->_lenptr = cur;

	return 1;
}

static int tlv_determine_len(struct tlv *t)
{
	unsigned char *cur = t->_lenptr;

	t->val = t->_lenptr+1;

	if (*cur & 0x80) {
		unsigned int len_len = *cur & 0x7f;

		t->val += len_len;

		/* we don't support length bigger 32bit */
		if (len_len > 4)
			return -ERANGE;

		/* make sure we don't read over end of buffer */
		if (cur + len_len > t->_tagptr + t->_len)
			return -EINVAL;

		if (len_len == 1)
			t->len = *(cur+1);
		else if (len_len == 2)
			t->len = ntohs(*((u_int16_t *)(cur+1)));
		else if (len_len == 3)
			t->len = ntohl(*((u_int32_t *)(cur+1)) & 0xffffff);
		else if (len_len == 4)
			t->len = ntohl(*((u_int32_t *)(cur+1)));
	} else
		t->len = *cur;

	return 1;
}

int
tlv_parse(const unsigned char *data, int len, struct tlv *tlv)
{
	int rc;

	memset(tlv, 0, sizeof(*tlv));

	tlv->_tagptr = (unsigned char *)data;
	tlv->_len = len;

	rc = tlv_determine_tag(tlv);
	if (rc < 0)
		return rc;

	rc = tlv_determine_len(tlv);
	if (rc < 0)
		return rc;

#if 0
	printf("TLV Tag = 0x%x (real: %x), Len = %u\n", tlv->tag, 
		tlv_raw_tag(tlv), tlv->len);
#endif
	if (tlv_gross_len(tlv) > len)
		return -EINVAL;

	return 1;
}

/* this really sucks.  MRTD specs use ASN.1 BER, but the specs
 * do not give the real tag value (e.g. '0') but the one including
 * all internal bits, e.g. 0x60. */
u_int16_t
tlv_raw_tag(const struct tlv *t)
{
	int tag_len = t->_lenptr - t->_tagptr;
	u_int16_t rawtag;

	if (tag_len == 2)
		rawtag = ntohs(*((u_int16_t *)t->_tagptr));
	else if (tag_len == 1)
		rawtag = *(t->_tagptr) & 0xff;

	return rawtag;
}
	
tlvlen_t tlv_gross_len(const struct tlv *t)
{
	return t->len + (t->val - t->_tagptr);
}

int
tlv_parse_first_child(const struct tlv *parent, struct tlv *tlv)
{
	return tlv_parse(parent->val, parent->len, tlv);	
}

int
tlv_parse_next_child(const struct tlv *parent, const struct tlv *prev, struct tlv *tlv)
{
	unsigned char *data = prev->val + prev->len;
	unsigned int remain_len = parent->len - (data - parent->val);

	if (data > parent->val + parent->len)
		return -EINVAL;

	return tlv_parse(data, remain_len, tlv);
}


int
tlv_build_tl(unsigned char *dst, unsigned int raw_tag, unsigned int len)
{
	unsigned int lenlen;
	unsigned char *cur = dst;
	/* we only support single byte tags */
	if (raw_tag > 0xff)
		return -EINVAL;

	*(cur++) = raw_tag & 0xff;

	if (len & 0xff000000)
		lenlen = 4;
	else if (len & 0x00ff0000)
		lenlen = 3;
	else if (len & 0x0000ff00)
		lenlen = 2;
	else
		lenlen = 1;

	if (lenlen == 1 && (len & 0x80) == 0)
		*(cur++) = len & 0x7f;
	else {
		*(cur++) = lenlen | 0x80;

		if (lenlen >= 4)
			*(cur++) = (len & 0xff000000) >> 24;
		if (lenlen >= 3)
			*(cur++) = (len & 0x00ff0000) >> 16;
		if (lenlen >= 2)
			*(cur++) = (len & 0x0000ff00) >> 8;

		*(cur++) = len & 0xff;
	}

	return (cur - dst);
}

/* build tlv */
int
tlv_build(unsigned char *dst, const unsigned char *src, unsigned int raw_tag, unsigned int len)
{
	int tl_len = tlv_build_tl(dst, raw_tag, len);

	memcpy(dst+tl_len, src, len);

	return 0;
}
